#include "../../kernel/system.h"

static inline uint8_t vgaEntryColor(enum vga_color fg, enum vga_color bg) {
	return fg | bg << 4;
}
 
static inline uint16_t vgaEntry(unsigned char uc, uint8_t color) {
	return (uint16_t) uc | (uint16_t) color << 8;
}

static void move_cursor()
{
    uint16_t cursorLocation = cursorY * 80 + cursorX;
    outb(0x3D4, 14);                  // Tell the VGA board we are setting the high cursor byte.
    outb(0x3D5, cursorLocation >> 8); // Send the high cursor byte.
    outb(0x3D4, 15);                  // Tell the VGA board we are setting the low cursor byte.
    outb(0x3D5, cursorLocation);      // Send the low cursor byte.
}

void scroll()
{
    if(cursorY >= 25)
    {
        int i;
        for (i = 0*80; i < 24*80; i++)
        {
            TBuffer[i] = TBuffer[i+80];
        }

        for (i = 24*80; i < 25*80; i++)
        {
            TBuffer[i] = ' ';
        }
        cursorY = 24;
    }
}
 
void monInit(void) {
	TRow = 0;
	TColumn = 0;
	TColor = vgaEntryColor(VGA_COLOR_LIGHT_GREY, VGA_COLOR_BLACK);
	TBuffer = (uint16_t*) 0xB8000;
	for (size_t y = 0; y < VGA_HEIGHT; y++) {
		for (size_t x = 0; x < VGA_WIDTH; x++) {
			const size_t index = y * VGA_WIDTH + x;
			TBuffer[index] = vgaEntry(' ', TColor);
		}
	}
    
    cursorX = 0;
    cursorY = 0;
    move_cursor();
}
 
void monSetColor(uint8_t color) {
	TColor = color;
}
 
void monPutEntryAt(char c, uint8_t color, size_t x, size_t y) {
	const size_t index = y * VGA_WIDTH + x;
    if (c == '\n') {
        if (TRow == VGA_HEIGHT-1) scroll();
        else TRow++;
        //c = '\0';
        TColumn = 0;
    }
	else TBuffer[index] = vgaEntry(c, color);
}
 
void monPutchar(char c) {
	monPutEntryAt(c, TColor, TColumn, TRow);
	if (++TColumn == VGA_WIDTH) {
		TColumn = 0;
		if (++TRow == VGA_HEIGHT) {
            scroll();
			TRow = 24;
        }
	}
}
 
void monPut(char c)
{
    // Handle a backspace
    if (c == 0x08 && cursorX)
    {
        TColumn = cursorX-1;
        TRow = cursorY;
        monPutchar(' ');
        cursorX--;
        if (cursorX == 0) {
            cursorX = 79;
            cursorY--;
        }
    }

    // Handle a tab
    else if (c == 0x09)
    {
        cursorX = (cursorX+8) & ~(8-1);
    }
    
    // Handle carriage return
    else if (c == '\r')
    {
        cursorX = 0;
    }

    // Handle newline
    else if (c == '\n')
    {
        cursorX = 0;
        cursorY++;
    }
    // Handle other printable character.
    else if(c >= ' ')
    {
        if (capsLock == 1 && (c >= 0x61 && c <= 0x7A)) {
            TColumn = cursorX;
            TRow = cursorY;
            monPutchar(c-32);
            cursorX++;
        }
        else {
            TColumn = cursorX;
            TRow = cursorY;
            monPutchar(c);
            cursorX++;
        }
    }

    if (cursorX >= 80)
    {
        cursorX = 0;
        cursorY ++;
    }

    scroll();
    move_cursor();

}

void monWrite(const char* data, size_t size) {
	for (size_t i = 0; i < size; i++)
//		monPutchar(data[i]);
        monPut(data[i]);
}
 
void monWString(const char* data) {
	monWrite(data, strlen(data));
}


void monWDec(uint32_t n)
{

    if (n == 0)
    {
        monPut('0');
        return;
    }

    int32_t acc = n;
    char c[32];
    int i = 0;
    while (acc > 0)
    {
        c[i] = '0' + acc%10;
        acc /= 10;
        i++;
    }
    c[i] = 0;

    char c2[32];
    c2[i--] = 0;
    int j = 0;
    while(i >= 0)
    {
        c2[i--] = c[j++];
    }
    monWString(c2);

}

void monWHex(uint32_t n)
{
    int32_t tmp;

    monWString("0x");

    char noZeroes = 1;

    int i;
    for (i = 28; i > 0; i -= 4)
    {
        tmp = (n >> i) & 0xF;
        if (tmp == 0 && noZeroes != 0)
        {
            continue;
        }
    
        if (tmp >= 0xA)
        {
            noZeroes = 0;
            monPut (tmp-0xA+'a' );
        }
        else
        {
            noZeroes = 0;
            monPut( tmp+'0' );
        }
    }
  
    tmp = n & 0xF;
    if (tmp >= 0xA)
    {
        monPut (tmp-0xA+'a');
    }
    else
    {
        monPut (tmp+'0');
    }

}